# -*- coding: utf-8 -*-
# @Time : 20-6-9 上午10:20
# @Author : zhuying
# @Company : Minivision
# @File : anti_spoof_predict.py
# @Software : PyCharm

import os
import cv2
import math
import torch
import numpy as np
import torch.nn.functional as F


from model_lib.MiniFASNet import MiniFASNetV1, MiniFASNetV2,MiniFASNetV1SE,MiniFASNetV2SE
from data_io import transform as trans
from utility import get_kernel, parse_model_name

MODEL_MAPPING = {
    'MiniFASNetV1': MiniFASNetV1,
    'MiniFASNetV2': MiniFASNetV2,
    'MiniFASNetV1SE':MiniFASNetV1SE,
    'MiniFASNetV2SE':MiniFASNetV2SE
}


class Detection:
    def __init__(self):
        caffemodel = "./detection_model/Widerface-RetinaFace.caffemodel"
        deploy = "./detection_model/deploy.prototxt"
        self.detector = cv2.dnn.readNetFromCaffe(deploy, caffemodel)
        self.detector_confidence = 0.6

    def get_bbox(self, img):
        height, width = img.shape[0], img.shape[1]
        aspect_ratio = width / height
        if img.shape[1] * img.shape[0] >= 192 * 192:
            img = cv2.resize(img,
                             (int(192 * math.sqrt(aspect_ratio)),
                              int(192 / math.sqrt(aspect_ratio))), interpolation=cv2.INTER_LINEAR)

        blob = cv2.dnn.blobFromImage(img, 1, mean=(104, 117, 123))
        self.detector.setInput(blob, 'data')
        out = self.detector.forward('detection_out').squeeze()
        max_conf_index = np.argmax(out[:, 2])
        left, top, right, bottom = out[max_conf_index, 3]*width, out[max_conf_index, 4]*height, \
                                   out[max_conf_index, 5]*width, out[max_conf_index, 6]*height
        bbox = [int(left), int(top), int(right-left+1), int(bottom-top+1)]
        return bbox


class AntiSpoofPredict(Detection):
    def __init__(self, device_id):
        super(AntiSpoofPredict, self).__init__()
        self.device = torch.device("cuda:{}".format(device_id)
                                   if torch.cuda.is_available() else "cpu")

    def _load_model(self, model_path):
        # define model
        model_name = os.path.basename(model_path)
        h_input, w_input, model_type, _ = parse_model_name(model_name)
        self.kernel_size = get_kernel(h_input, w_input,)
        self.model = MODEL_MAPPING[model_type](conv6_kernel=self.kernel_size).to(self.device)

        # load model weight
        state_dict = torch.load(model_path, map_location=self.device)
        keys = iter(state_dict)
        first_layer_name = keys.__next__()
        if first_layer_name.find('module.') >= 0:
            from collections import OrderedDict
            new_state_dict = OrderedDict()
            for key, value in state_dict.items():
                name_key = key[7:]
                new_state_dict[name_key] = value
            self.model.load_state_dict(new_state_dict)
        else:
            self.model.load_state_dict(state_dict)
        return None

    def predict(self, img, model_path):
        img_transform = trans.Compose([
            trans.ToTensor(),
        ])
        img = img_transform(img)
        img = img.unsqueeze(0).to(self.device)
        self._load_model(model_path)
        self.model.eval()
        with torch.no_grad():
            result = self.model.forward(img)
            result = F.softmax(result).cpu().numpy()
        return result














# -*- coding: utf-8 -*-
# @Time : 20-6-9 下午3:06
# @Author : zhuying
# @Company : Minivision
# @File : test.py
# @Software : PyCharm

import os
import cv2
import numpy as np
import argparse
import warnings
import time

from anti_spoof_predict import AntiSpoofPredict
from generate_patches import CropImage
from utility import parse_model_name
warnings.filterwarnings('ignore')


SAMPLE_IMAGE_PATH = "./images/"


# 因为安卓端APK获取的视频流宽高比为3:4,为了与之一致，所以将宽高比限制为3:4
class Liveness():
    def check_image(self,image):
        # import pdb;pdb.set_trace()
        # height, width, channel = image.shape
        # if width/height:
        #     print("Image is not appropriate!!!\nHeight/Width should be 4/3.")
        #     return False
        # else:
        return True


    def test(self,image, model_dir='./anti_spoof_models',device_id=0):
        model_test = AntiSpoofPredict(device_id)
        image_cropper = CropImage()
        width = int((image.shape[0]*3)/4)
        dim=(width,image.shape[0])
        resized = cv2.resize(image, dim, interpolation = cv2.INTER_AREA)
        result = self.check_image(resized)
        if result is False:
            return
        image_bbox = model_test.get_bbox(image)
        prediction = np.zeros((1, 3))
        test_speed = 0
        # sum the prediction from single model's result
        for model_name in os.listdir(model_dir):
            h_input, w_input, model_type, scale = parse_model_name(model_name)
            param = {
                "org_img": image,
                "bbox": image_bbox,
                "scale": scale,
                "out_w": w_input,
                "out_h": h_input,
                "crop": True,
            }
            if scale is None:
                param["crop"] = False
            img = image_cropper.crop(**param)
            start = time.time()
            prediction += model_test.predict(img, os.path.join(model_dir, model_name))
            test_speed += time.time()-start

        # draw result of prediction
        label = np.argmax(prediction)
        value = prediction[0][label]/2
        if label == 1:
            res={"score":value,
                "result":"real",
                "rectangle":{
                "left":image_bbox[0],
                "top":image_bbox[1],
                "right":image_bbox[0] + image_bbox[2],
                "bottom":image_bbox[1] + image_bbox[3]
                }}
            # print("Image '{}' is Real Face. Score: {:.2f}.".format(image_name, value))
            result_text = "RealFace Score: {:.2f}".format(value)
            color = (255, 0, 0)
        else:
            res={"score":value,
                "result":"fake",
                "rectangle":{
                "left":image_bbox[0],
                "top":image_bbox[1],
                "right":image_bbox[0] + image_bbox[2],
                "bottom":image_bbox[1] + image_bbox[3]
                }}
            # print("Image '{}' is Fake Face. Score: {:.2f}.".format(image_name, value))
            result_text = "FakeFace Score: {:.2f}".format(value)
            color = (0, 0, 255)
        print("Prediction cost {:.2f} s".format(test_speed))
        # cv2.rectangle(
        #     image,
        #     (image_bbox[0], image_bbox[1]),
        #     (image_bbox[0] + image_bbox[2], image_bbox[1] + image_bbox[3]),
        #     color, 2)
        # cv2.putText(
        #     image,
        #     result_text,
        #     (image_bbox[0], image_bbox[1] - 5),
        #     cv2.FONT_HERSHEY_COMPLEX, 0.5*image.shape[0]/1024, color)

        # format_ = os.path.splitext(image_name)[-1]
        # result_image_name = image_name.replace(format_, "_result" + format_)
        # cv2.imwrite(SAMPLE_IMAGE_PATH + result_image_name, image)

        return res


    # if __name__ == "__main__":
    #     desc = "test"
    #     parser = argparse.ArgumentParser(description=desc)
    #     parser.add_argument(
    #         "--device_id",
    #         type=int,
    #         default=0,
    #         help="which gpu id, [0/1/2/3]")
    #     parser.add_argument(
    #         "--model_dir",
    #         type=str,
    #         default="./anti_spoof_models",
    #         help="model_lib used to test")
    #     parser.add_argument(
    #         "--image_name",
    #         type=str,
    #         default="image_F1.jpg",
    #         help="image used to test")
    #     args = parser.parse_args()
    #     print(args)
    result=test(image_name='mihirdai.jpg')
    print(result)



# -*- coding: utf-8 -*-
# @Time : 20-6-4 下午2:13
# @Author : zhuying
# @Company : Minivision
# @File : utility.py
# @Software : PyCharm

from datetime import datetime
import os


def get_time():
    return (str(datetime.now())[:-10]).replace(' ', '-').replace(':', '-')


def get_kernel(height, width):
    kernel_size = ((height + 15) // 16, (width + 15) // 16)
    return kernel_size


def get_width_height(patch_info):
    w_input = int(patch_info.split('x')[-1])
    h_input = int(patch_info.split('x')[0].split('_')[-1])
    return w_input,h_input


def parse_model_name(model_name):
    info = model_name.split('_')[0:-1]
    h_input, w_input = info[-1].split('x')
    model_type = model_name.split('.pth')[0].split('_')[-1]

    if info[0] == "org":
        scale = None
    else:
        scale = float(info[0])
    return int(h_input), int(w_input), model_type, scale


def make_if_not_exist(folder_path):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)



from distutils.core import setup
setup(
  name = 'image-liveness',         
  packages = ['Liveness'],  
  version = '1.2',      
  license='Apache License 2.0',        
  description = 'This library offers a liveness detection on static images, which can be used to distinguish if it is a live capture or a spoofed image based on facial features and other indicators of authenticity.',  
  author = 'Ashish Thapa',                   
  author_email = 'ashish.f1soft@gmail.com',      
  url = 'https://github.com/ashishf1soft',  
  download_url = 'https://github.com/ashishf1soft/liveness-detection/archive/refs/tags/v_01.tar.gz',   
  keywords = ['LIVENESS','FACE FACE','DETECTION'], 
  install_requires=[            # I get to this in a second
          'numpy',
          'opencv-python',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: Apache Software License',   
    'Programming Language :: Python :: 3.9',
  ],
)